@@ -30,7 +30,6 @@
 
 
 /* --- defines --- */
-#define G_PARAM_SPEC_CLASS(class)		(G_TYPE_CHECK_CLASS_CAST ((class), G_TYPE_PARAM, GParamSpecClass))
 #define	G_PARAM_USER_MASK			(~0 << G_PARAM_USER_SHIFT)
 #define PSPEC_APPLIES_TO_VALUE(pspec, value)	(G_TYPE_CHECK_VALUE_TYPE ((value), G_PARAM_SPEC_VALUE_TYPE (pspec)))
 #define	G_SLOCK(mutex)				g_static_mutex_lock (mutex)
@@ -743,7 +742,7 @@ pool_list (gpointer key,
 {
   GParamSpec *pspec = value;
   gpointer *data = user_data;
-  GType owner_type = GPOINTER_TO_UINT (data[1]);
+  GType owner_type = (GType) data[1];
 
   if (owner_type == pspec->owner_type)
     data[0] = g_list_prepend (data[0], pspec);
@@ -760,7 +759,7 @@ g_param_spec_pool_list_owned (GParamSpecPool *pool,
   
   G_SLOCK (&pool->smutex);
   data[0] = NULL;
-  data[1] = GUINT_TO_POINTER (owner_type);
+  data[1] = (gpointer) owner_type;
   g_hash_table_foreach (pool->hash_table, pool_list, &data);
   G_SUNLOCK (&pool->smutex);
 
@@ -810,7 +809,7 @@ pool_depth_list (gpointer key,
   GParamSpec *pspec = value;
   gpointer *data = user_data;
   GSList **slists = data[0];
-  GType owner_type = GPOINTER_TO_UINT (data[1]);
+  GType owner_type = (GType) data[1];
 
   if (g_type_is_a (owner_type, pspec->owner_type))
     {
@@ -839,7 +838,7 @@ g_param_spec_pool_list (GParamSpecPool *pool,
   d = g_type_depth (owner_type);
   slists = g_new0 (GSList*, d);
   data[0] = slists;
-  data[1] = GUINT_TO_POINTER (owner_type);
+  data[1] = (gpointer) owner_type;
   g_hash_table_foreach (pool->hash_table, pool_depth_list, &data);
   for (i = 0; i < d - 1; i++)
     slists[i] = pspec_list_remove_overridden (slists[i], pool->hash_table, owner_type, n_pspecs_p);
@@ -963,6 +962,19 @@ g_value_set_param (GValue     *value,
     g_param_spec_ref (value->data[0].v_pointer);
 }
 
+void
+g_value_set_param_take_ownership (GValue     *value,
+				  GParamSpec *param)
+{
+  g_return_if_fail (G_VALUE_HOLDS_PARAM (value));
+  if (param)
+    g_return_if_fail (G_IS_PARAM_SPEC (param));
+
+  if (value->data[0].v_pointer)
+    g_param_spec_unref (value->data[0].v_pointer);
+  value->data[0].v_pointer = param; /* we take over the reference count */
+}
+
 GParamSpec*
 g_value_get_param (const GValue *value)
 {
