@@ -37,6 +37,8 @@
 
 #include <glib.h>
 
+#ifdef G_THREADS_ENABLED
+
 static gboolean thread_system_already_initialized = FALSE;
 static gint g_thread_priority_map [G_THREAD_PRIORITY_URGENT + 1];
 
@@ -274,6 +276,19 @@ g_thread_init_with_errorcheck_mutexes (GThreadFunctions* init)
   if (init)
     g_error ("Errorcheck mutexes can only be used for native " 
 	     "thread implementations. Sorry." );
+
+#ifdef HAVE_G_THREAD_IMPL_INIT
+  /* This isn't called in g_thread_init, as it doesn't think to get
+   * the default implementation, so we have to call it on our own.
+   *
+   * We must call this before copying
+   * g_thread_functions_for_glib_use_default as the
+   * implementation-specific init function might modify the contents
+   * of g_thread_functions_for_glib_use_default based on operating
+   * system version, C library version, or whatever. */
+  g_thread_impl_init();
+#endif /* HAVE_G_THREAD_IMPL_INIT */
+
   errorcheck_functions = g_thread_functions_for_glib_use_default;
   errorcheck_functions.mutex_new = g_mutex_new_errorcheck_impl;
   errorcheck_functions.mutex_lock = 
@@ -290,12 +305,6 @@ g_thread_init_with_errorcheck_mutexes (GThreadFunctions* init)
     (gboolean (*)(GCond *, GMutex *, GTimeVal *)) 
     g_cond_timed_wait_errorcheck_impl;
     
-#ifdef HAVE_G_THREAD_IMPL_INIT
-  /* This isn't called in g_thread_init, as it doesn't think to get
-   * the default implementation, so we have to call it on our own. */
-  g_thread_impl_init();
-#endif
-
   g_thread_init (&errorcheck_functions);
 }
 
@@ -304,10 +313,6 @@ g_thread_init (GThreadFunctions* init)
 {
   gboolean supported;
 
-#ifndef	G_THREADS_ENABLED
-  g_error ("GLib thread support is disabled.");
-#endif	/* !G_THREADS_ENABLED */
-
   if (thread_system_already_initialized)
     g_error ("GThread system may only be initialized once.");
     
@@ -383,3 +388,13 @@ g_thread_init (GThreadFunctions* init)
   /* we want the main thread to run with normal priority */
   g_thread_set_priority (g_thread_self(), G_THREAD_PRIORITY_NORMAL);
 }
+
+#else /* !G_THREADS_ENABLED */
+
+void
+g_thread_init (GThreadFunctions* init)
+{
+  g_error ("GLib thread support is disabled.");
+}
+
+#endif /* !G_THREADS_ENABLED */
